package cat.hack3.codingtests.marsrover.gps;

import cat.hack3.codingtests.marsrover.api.cartography.Coordinates;
import cat.hack3.codingtests.marsrover.api.cartography.Direction;
import cat.hack3.codingtests.marsrover.api.cartography.PlanetMap;

import java.util.List;

/**
 * Notice that the representation is the following:
 * Latitude is Y, starts from 1 on the top and goes down incrementally
 * Longitude is X, starts from 1 on the left and goes to right incrementally
 *     1    2    3
 * 1  1-1  1-2  1-3
 * 2  2-1  2-2  2-3
 * 3  3-1  3-2  3-3
 */
public class MarsMap implements PlanetMap {

    private static final int FIRST_POSITION_IN_MAP = 1;
    private static final int INCREMENT_UNIT = 1;

    private Coordinates currentPosition;
    private final List<Coordinates> obstaclesLocalizations;

    private final MapIncrementalPositionResolver positionResolver;

    public static class Provider implements PlanetMap.Provider {
        @Override
        public PlanetMap provideWith(int height, int width, Coordinates startingCoordinates, List<Coordinates> obstaclesLocalizations) {
            return new MarsMap(height, width, startingCoordinates, obstaclesLocalizations);
        }
    }

    private MarsMap(int height, int width, Coordinates startingCoordinates, List<Coordinates> obstaclesLocalizations) {
        currentPosition = startingCoordinates;
        this.obstaclesLocalizations = obstaclesLocalizations;
        positionResolver = new MapIncrementalPositionResolver(FIRST_POSITION_IN_MAP, INCREMENT_UNIT, height, width);
    }

    @Override
    public Coordinates getNextPositionTowards(Direction direction) {
        return switch (direction) {
            case NORTH -> getDecrementedLatitude();
            case SOUTH -> getIncrementedLatitude();
            case WEST -> getDecrementedLongitude();
            case EAST -> getIncrementedLongitude();
        };
    }

    private Coordinates getIncrementedLatitude() {
        var latitude = currentPosition.latitude();
        var newLatitude = positionResolver.getIncremented(latitude);
        return currentPosition.withUpdated(newLatitude);
    }

    private Coordinates getDecrementedLatitude() {
        var latitude = currentPosition.latitude();
        var newLatitude = positionResolver.getDecremented(latitude);
        return currentPosition.withUpdated(newLatitude);
    }

    private Coordinates getIncrementedLongitude() {
        var longitude = currentPosition.longitude();
        var newLongitude = positionResolver.getIncremented(longitude);
        return currentPosition.withUpdated(newLongitude);
    }

    private Coordinates getDecrementedLongitude() {
        var longitude = currentPosition.longitude();
        var newLongitude = positionResolver.getDecremented(longitude);
        return currentPosition.withUpdated(newLongitude);
    }

    @Override
    public synchronized void updateLocation(Coordinates newPosition) {
        currentPosition = newPosition;
    }

    @Override
    public boolean willCollideWithObstacle(Coordinates newCoordinates) {
        return obstaclesLocalizations.contains(newCoordinates);
    }

    @Override
    public Coordinates getCurrentPosition() {
        return currentPosition;
    }
}
